class UserReadableException(Exception):
    pass


class SecretDoesNotExists(UserReadableException):
    pass


class DatabaseEngineNotSupported(UserReadableException):
    pass


class DatabaseCreationError(UserReadableException):
    pass


class ServiceDoesNotExists(UserReadableException):
    pass


class TaskDoesNotExists(UserReadableException):
    pass


class SecretCreationError(UserReadableException):
    pass


class SecretRemovalError(UserReadableException):
    pass


class ContainerDoesNotExists(UserReadableException):
    pass


class SpaceCreationError(UserReadableException):
    pass



import logging
import sys

import click

from edos import commands

LOG = logging.getLogger(__name__)

TRACE = logging.DEBUG - 1
logging.addLevelName(TRACE, "TRACE")


@click.group()
@click.option("--verbose", is_flag=True)
@click.version_option("1.13.0", "--version", "-v")
def main(verbose):
    from edos.exceptions import UserReadableException

    # Set verbosity level
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        format="%(asctime)s,%(msecs)03d - %(message)s",
        level=log_level,
        datefmt="%H:%M:%S",
    )

    def exception_handler(type, value, traceback):
        if verbose:
            return old_excepthook(type, value, traceback)

        if isinstance(value, UserReadableException):
            LOG.error(str(value))
        elif hasattr(value, "body"):
            # Kubernetes API exception have the interesting info in .body
            # but we avoid importing Kubernetes API to speed up start (200ms).
            LOG.error(getattr(value, "body", str(value)))
        else:
            return old_excepthook(type, value, traceback)

    old_excepthook = sys.excepthook
    sys.excepthook = exception_handler


commands.register_commands(main)
if __name__ == "__main__":
    main()



import json
import logging
import math
import os
from json import JSONDecodeError

import click
from tabulate import tabulate
import yaml

from edos.cache import cache
from edos.cli.common import create_completion_from_names, get_new_secret_name
from edos.exceptions import SecretDoesNotExists
from edos.services.docker_service import DockerService
from edos.services.swarmpit_service import SwarmpitService
from edos.settings import conf

LOG = logging.getLogger(__name__)


@click.group()
def docker():
    """Managing our docker swarm cluster"""


def service_completion(ctx, args, incomplete):
    @cache.memoize(expire=10)
    def get_names():
        return [s.name for s in DockerService().get_services()]

    return create_completion_from_names(get_names(), incomplete)


def secrets_completion(ctx, args, incomplete):
    @cache.memoize(expire=10)
    def get_names():
        return [s.name for s in DockerService().get_secrets()]

    return create_completion_from_names(get_names(), incomplete)


@docker.command()
@click.argument("service_name", shell_complete=service_completion)
@click.option(
    "--command",
    help="Specify which command will be used [bash, sh etc.]",
    type=str,
)
def exec(service_name, command):
    """\"docker exec\" into a service container"""
    if not command:
        command = "bash"
    d_service = DockerService()
    click.secho("Retrieving hostname...", fg="blue")
    host_name = d_service.get_node_name_for_service(service_name)
    click.secho("Retrieving container...", fg="blue")
    container_name = d_service.get_container(host_name, service_name)
    click.secho("Connecting to server...", fg="blue")
    os.system(f"docker --host ssh://{host_name} exec -it {container_name} {command}")

@docker.command()
@click.argument("service_name", shell_complete=service_completion)
@click.option("--format", type=click.Choice(["json", "yaml"]), default="json")
def inspect(service_name, format):
    """Inspect a service"""
    d_service = DockerService()
    service = d_service.get_service_by_service_name(service_name)
    if format == "json":
        click.echo(json.dumps(service.attrs, indent=2))
    else:
        click.echo(yaml.dump(service.attrs))

@docker.group()
def secrets():
    """Managing docker secrets"""


@secrets.command()
def ls():
    """List secrets"""
    items = DockerService().get_secrets()
    for secret in items:
        click.echo(secret.name)


@secrets.command()
@click.option("--name", type=str, help="Name of the secret", required=False)
@click.option("--from-file", type=click.Path(exists=True), help="File to read secret from", required=False)
@click.option("--no-confirm", is_flag=True, help="Do not ask for confirmation")
def create(name, from_file, no_confirm = False):
    """Create a secret interactively"""
    if not name:
        name = click.prompt("Name")
    if not no_confirm:
        click.confirm(
            "An editor will be opened to input the secret's value. " "Exit without saving to abort.",
            default=True,
            show_default=False,
            prompt_suffix="",
        )
    if from_file:
        with open(from_file, "rb") as f:
            value = f.read()
    else:
        value = click.edit()
    if name and value:
        result = DockerService().create_secret(name, value)
        click.echo(click.style("Created secret with ID: ", fg="green") + result.id)
    else:
        click.secho(f"Creation of secret '{name}' cancelled", fg="red")


@secrets.command()
@click.argument("secret_name", type=str, shell_complete=secrets_completion)
@click.option("--no-pager", is_flag=True, help="Do not use pager")
def ps(secret_name: str, no_pager: bool = False):
    """Read one secret"""
    d_service = DockerService()
    click.secho("Getting secret", fg="blue")
    d_secret = d_service.get_secret_by_name(secret_name)
    click.secho("Creating container with mounted secret", fg="blue")
    _service = d_service.create_service_with_secret(d_secret)
    try:
        container_name = d_service.get_container_name_by_service(_service)

        command = f"docker --host {conf.DOCKER_SSH_BASE_URL} exec -t {container_name} " f"cat /run/secrets/{d_secret.name}"

        old_secret = os.popen(command).read()
        if no_pager:
            click.echo("=" * 34)
            click.echo(old_secret)
            click.echo("=" * 34)
        else:
            click.echo_via_pager(old_secret)
    finally:
        _service.remove()


@secrets.command()
@click.argument("secret_name", type=str, shell_complete=secrets_completion)
def rm(secret_name: str):
    """Remove a secret"""
    DockerService().remove_secret(secret_name)
    click.secho("Secret has been removed", fg="green")


@secrets.command()
@click.argument("secret_name", type=str, shell_complete=secrets_completion)
def edit(secret_name: str):
    """
    1. check if secret exists
    2. create container and mount secret to it
    3. open and edit secret
    4. create new secret and return secret name
    :param secret_name:
    :return:
    """
    d_service = DockerService()
    click.secho("Getting secret", fg="blue")
    d_secret = d_service.get_secret_by_name(secret_name)
    click.secho("Creating container with mounted secret", fg="blue")
    _service = d_service.create_service_with_secret(d_secret)
    try:
        container_name = d_service.get_container_name_by_service(_service)

        command = f"docker --host {conf.DOCKER_SSH_BASE_URL} exec -t {container_name} " f"cat /run/secrets/{d_secret.name}"
        click.secho("Getting secret content", fg="blue")
        old_secret = os.popen(command).read()

        while True:
            new_data = click.edit(old_secret, extension=".json")
            if new_data and old_secret.strip() != new_data.strip():
                break
            if not click.confirm("No change detected, do you want to edit again?"):
                return click.secho("Creation cancelled", fg="red")
            continue

        try:
            json.loads(new_data)
        except JSONDecodeError:
            if not click.confirm("Not valid JSON, do you want to continue?"):
                return click.secho("Creation cancelled", fg="red")

        new_name = get_new_secret_name(secret_name)

        while True:
            name = click.prompt(
                "New secret name: (keep empty if you want to use default)",
                default=new_name,
            )
            try:
                d_service.get_secret_by_name(name)
                click.secho("Secret already exists, choose another name", fg="yellow")
            except SecretDoesNotExists:
                break

        result = DockerService().create_secret(name, bytes(new_data.encode("utf-8")))
        click.echo(click.style("Created secret with name: ", fg="green") + result.name)
    finally:
        _service.remove()


@docker.group()
def service():
    """Managing swarm services"""


@service.command()
# flake8: noqa
def ls():  # flake8: noqa
    """List running services"""
    for item in DockerService().get_services():
        click.echo(item.name)


@service.command()
def reservation():
    """List of services without reservation"""
    services = DockerService().get_services_without_reservation()
    if not services:
        click.echo("Good job. All services has reservation limit")
    click.echo("These services have no resource limits:")
    for s in services:
        click.secho(s.name, fg="red")


def get_mem_res_text(res_memory, avg_memory, diff):
    if not res_memory:
        return click.style("No limit", fg="red")

    if res_memory < avg_memory:
        return click.style(res_memory, fg="red")

    if res_memory - diff < avg_memory:
        return click.style(res_memory, fg="yellow")

    return click.style(res_memory, fg="green")


sort_cols = ["name", "memory"]
sort_mapper = {val: i for i, val in enumerate(sort_cols)}


@service.command()
@click.option(
    "--diff",
    help="Specify how many extra MiB services can have (else warning)",
    type=int,
)
@click.option(
    "--sort",
    help="Specify sorting",
    type=click.Choice(sort_cols),
    default=sort_cols[1],
)
def memstats(diff, sort):
    """List of memory stats for all services"""
    memory_stats = SwarmpitService().get_service_memory_stats()
    if not diff:
        diff = 20

    text_to_tabulate = []
    for s in memory_stats:
        res_text = get_mem_res_text(s.memory_reservation, s.actual_memory_usage, diff)
        text_to_tabulate.append([s.service_name, s.actual_memory_usage, res_text])

    text_to_tabulate = sorted(text_to_tabulate, key=lambda x: x[sort_mapper[sort]], reverse=True)

    click.echo(tabulate(text_to_tabulate, headers=["Service", "Usage", "Reservation"]))


def convert_size(size_bytes):
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])


@service.command()
def reservationsum():
    sum_res = DockerService().get_sum_of_reservation()
    click.echo(f"Sum of reservation: {convert_size(sum_res)}")



import datetime
import uuid

from clockify_api_client.client import ClockifyAPIClient
from clockify_api_client.models.report import Report
from clockify_api_client.models.user import User
from dateutil.relativedelta import relativedelta
from fakturoid import Fakturoid, Invoice, InvoiceLine

from edos.exceptions import UserReadableException
from edos.settings import conf


class InvoiceService:
    def __init__(self):
        self.fakturoid = Fakturoid(
            conf.INVOICE_CONFIG.get("fakturoid_account_slug"),
            conf.INVOICE_CONFIG.get("fakturoid_account_email"),
            conf.INVOICE_CONFIG.get("fakturoid_api_key"),
            conf.FAKTUROID_USER_AGENT,
        )
        self.clockify = ClockifyAPIClient().build(conf.INVOICE_CONFIG.get("clockify_api_key"), conf.CLOCKIFY_API_URL)

    def create_invoice(self) -> str:
        """generate invoice and return its URL"""
        last_month_hour_report = self.get_last_month_hour_report()
        custom_id = uuid.uuid4()
        invoice = Invoice(
            subject_id=self.get_subject_id(),
            issued_on=(datetime.datetime.now().replace(day=1) - relativedelta(days=1)).date(),
            lines=[
                InvoiceLine(
                    name="software development",
                    unit_name="hod",
                    quantity=last_month_hour_report,
                    unit_price=1,
                )
            ],
            custom_id=str(custom_id),
        )
        self.fakturoid.save(invoice)
        invoices = self.fakturoid.invoices(custom_id=custom_id)
        if len(invoices) == 0:
            raise UserReadableException("Invoice not found")
        return invoices[0].html_url

    def get_subject_id(self) -> int:
        for subject in self.fakturoid.subjects():
            if subject.registration_no == conf.ENDEVEL_ICO:
                return subject.id
        raise UserReadableException("Subject not found")

    def get_last_month_hour_report(self) -> int:
        report_api: Report = self.clockify.reports
        last_month_start = (datetime.datetime.now() - relativedelta(months=1)).replace(
            day=1, hour=0, minute=0, second=0, microsecond=0
        )
        last_month_end = (datetime.datetime.now().replace(day=1) - relativedelta(days=1)).replace(
            hour=23, minute=59, second=59, microsecond=999999
        )
        return (
            report_api.get_summary_report(
                conf.CLOCKIFY_ENDEVEL_WORKSPACE_ID,
                {
                    "dateRangeStart": last_month_start.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
                    "dateRangeEnd": last_month_end.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
                    "summaryFilter": {"groups": ["USER_GROUP"]},
                    "users": {
                        "ids": [self.get_clockify_user_id()],
                        "status": "ACTIVE_WITH_PENDING",
                        "contains": "CONTAINS",
                    },
                    "amountShown": "HIDE_AMOUNT",
                },
            )
            .get("totals", [])[0]
            .get("totalTime")
            / 3600
        )

    def get_clockify_user_id(self) -> str:
        user_api: User = self.clockify.users
        return user_api.get_current_user().get("id")



import configparser
import os

import click

from edos.settings import conf


def configure(
    token,
    aws_access_key,
    aws_secret_key,
    swarmpit_token,
    fakturoid_token,
    clockify_token,
):
    conf.DIGITAL_OCEAN_CONFIG = {"token": token}
    conf.AWS_CONFIG = {
        "aws_access_key": aws_access_key,
        "aws_secret_key": aws_secret_key,
    }
    conf.SWARMPIT_CONFIG = {"token": swarmpit_token}
    conf.INVOICE_CONFIG = {
        "faturoid_api_key": fakturoid_token,
        "clockify_api_key": clockify_token,
    }


def get_do_token() -> str:
    previous_token = conf.DIGITAL_OCEAN_CONFIG.get("token", "")
    click.echo(
        (
            "This tool needs your DigitalOcean API token for:\n"
            "    - Creating databases\n"
            "\n"
            "You can get your access token at:\n"
            "    https://cloud.digitalocean.com/settings/api/tokens"
        )
    )
    if previous_token:
        click.echo((f"Your current token: {previous_token}\n" "(Enter empty string to use your current token)"))
    token = ""
    while not token:
        token = click.prompt(
            click.style("Your DO token", fg="yellow"),
            default=previous_token,
            show_default=False,
            type=str,
        )
    return token


def get_aws_keys() -> tuple[str, str]:
    previous_aws_access_key = conf.AWS_CONFIG.get("aws_access_key", "")
    previous_aws_secret_key = conf.AWS_CONFIG.get("aws_secret_key", "")
    click.echo(("This tool needs your AWS AccessKey and SecretKey for:\n" "    - Creating spaces"))
    if previous_aws_access_key:
        click.echo(
            (f"Your current access_key: {previous_aws_access_key}\n" "(Enter empty string to use your current token)")
        )
    aws_access_key = ""
    while not aws_access_key:
        aws_access_key = click.prompt(
            click.style("Your AWS access key", fg="yellow"),
            default=previous_aws_access_key,
            show_default=False,
            type=str,
        )
    if previous_aws_secret_key:
        click.echo(
            (f"Your current secret_key: {previous_aws_secret_key}\n" "(Enter empty string to use your current token)")
        )
    aws_secret_key = ""
    while not aws_secret_key:
        aws_secret_key = click.prompt(
            click.style("Your AWS secret key", fg="yellow"),
            default=previous_aws_secret_key,
            show_default=False,
            type=str,
        )
    return aws_access_key, aws_secret_key


def get_swarmpit_token() -> str:
    previous_token = conf.SWARMPIT_CONFIG.get("token", "")
    click.echo(
        (
            "This tool needs your Swarmpit bearer API token for:\n"
            "    - Getting stats services (memory and reservation)\n"
            "\n"
            "You can get your access token at:\n"
            "    https://swarmpit.doc.endevel.cz/#/account-settings "
            "      or in our Lastpass\n"
            "(paste without Bearer keyword)"
        )
    )
    if previous_token:
        click.echo((f"Your current token: {previous_token}\n" "(Enter empty string to use your current token)"))
    token = ""
    while not token:
        token = click.prompt(
            click.style("Your Swarmpit token", fg="yellow"),
            default=previous_token,
            show_default=False,
            type=str,
        )
    return token


def prompt_function(prompt_message: str, previous_token: str) -> str:
    if previous_token:
        click.echo((f"Your current token: {previous_token}\n" "(Enter empty string to use your current token)"))
    token = ""
    while not token:
        token = click.prompt(
            click.style(prompt_message, fg="yellow"),
            default=previous_token,
            show_default=False,
            type=str,
        )
    return token


def get_fakturoid_information() -> tuple[str, str, str]:
    click.echo(
        "For generating invoices, this tool needs your:\n"
        "    - Fakturoid API key\n"
        "    - Fakturoid account email\n"
        "    - Fakturoid account slug (when you log-in, you see "
        'the slug in the url, "stepanbinko" for example)\n'
    )
    api_key = prompt_function("Your Fakturoid API key", conf.INVOICE_CONFIG.get("fakturoid_api_key", ""))
    account_email = prompt_function(
        "Your Fakturoid account email",
        conf.INVOICE_CONFIG.get("fakturoid_account_email", ""),
    )
    account_slug = prompt_function(
        "Your Fakturoid account slug",
        conf.INVOICE_CONFIG.get("fakturoid_account_slug", ""),
    )
    return api_key, account_email, account_slug


def get_clockify_token() -> str:
    previous_token = conf.INVOICE_CONFIG.get("clockify_api_key", "")
    click.echo("For generating invoices, this tool needs your clockify API key:\n")
    if previous_token:
        click.echo((f"Your current token: {previous_token}\n" "(Enter empty string to use your current token)"))
    token = ""
    while not token:
        token = click.prompt(
            click.style("Your clockify token", fg="yellow"),
            default=previous_token,
            show_default=False,
            type=str,
        )
    return token


def configure_interactive():
    click.echo(click.style("Welcome to the EDOS configuration tool", bold=True))

    click.echo()

    config_file = configparser.ConfigParser()
    config_file.add_section("DO")
    token = get_do_token()
    config_file.set("DO", "token", token)

    click.echo()

    config_file.add_section("AWS")
    aws_access_key, aws_secret_key = get_aws_keys()
    config_file.set("AWS", "aws_access_key", aws_access_key)
    config_file.set("AWS", "aws_secret_key", aws_secret_key)

    click.echo()

    config_file.add_section("SWARMPIT")
    token = get_swarmpit_token()
    config_file.set("SWARMPIT", "token", token)

    click.echo()
    config_file.add_section("INVOICE")
    (
        fakturoid_api_key,
        fakturoid_account_email,
        fakturoid_account_slug,
    ) = get_fakturoid_information()
    clockify_token = get_clockify_token()
    config_file.set("INVOICE", "fakturoid_api_key", fakturoid_api_key)
    config_file.set("INVOICE", "fakturoid_account_email", fakturoid_account_email)
    config_file.set("INVOICE", "fakturoid_account_slug", fakturoid_account_slug)
    config_file.set("INVOICE", "clockify_api_key", clockify_token)

    click.echo(f"Writing configuration file: {conf.CONFIG_PATH}")
    os.makedirs(conf.USER_DIR.user_config_dir, exist_ok=True)
    with open(conf.CONFIG_PATH, "w") as git_configfile:
        config_file.write(git_configfile)
    conf.reload()

    click.echo(click.style(">> Done <<", fg="green"))



import dataclasses
from datetime import datetime
from typing import Optional

from edos.api.digitalocean.databases.models.connections import DatabaseConnection
from edos.api.digitalocean.databases.models.database import Database
from edos.api.digitalocean.databases.models.database_user import DatabaseUser


@dataclasses.dataclass
class DatabaseCluster:
    id: str
    name: str
    engine: str
    version: str
    num_nodes: int
    size: str
    region: str
    status: str
    created_at: datetime
    private_network_uuid: str
    tags: Optional[list[str]]
    db_names: Optional[list[Database]]
    connection: DatabaseConnection
    privateConnection: DatabaseConnection
    users: Optional[list[DatabaseUser]]



from edos.api.digitalocean.base_api import DigitalOceanApi
from edos.api.digitalocean.databases.helpers import parse_one_cluster, parse_one_database, parse_one_user
from edos.api.digitalocean.databases.models.cluster import DatabaseCluster
from edos.api.digitalocean.databases.models.database import Database
from edos.api.digitalocean.databases.models.database_user import DatabaseUser


class DatabaseApi(DigitalOceanApi):
    path = "/v2/databases"

    def get_clusters(self) -> list[DatabaseCluster]:
        url = self.build_url()
        resp = self.get_request(url)["databases"]
        clusters = []
        for item in resp:
            clusters.append(parse_one_cluster(item))
        return clusters

    def get_one_cluster(self, cluster_id: str) -> DatabaseCluster:
        cluster_path = f"/{cluster_id}"
        url = self.build_url() + cluster_path
        resp = self.get_request(url)["database"]
        return parse_one_cluster(resp)

    def get_databases(self, cluster_id: str) -> list[Database]:
        cluster_path = f"/{cluster_id}/dbs"
        url = self.build_url() + cluster_path
        resp = self.get_request(url)["dbs"]
        databases = []
        for item in resp:
            databases.append(parse_one_database(item))
        return databases

    def get_database(self, cluster_id: str, database_name: str) -> Database:
        url = self.build_url() + f"/{cluster_id}/dbs/{database_name}"
        resp = self.get_request(url)["db"]
        print(resp)
        return parse_one_database(resp)

    def create_database(self, cluster_id: str, database_name: str) -> Database:
        cluster_path = f"/{cluster_id}/dbs"
        url = self.build_url() + cluster_path
        payload = {
            "name": database_name,
        }
        resp = self.post_request(url, payload)
        return parse_one_database(resp["db"])

    def create_database_user(self, cluster_id: str, user_name: str) -> DatabaseUser:
        cluster_path = f"/{cluster_id}/users"
        url = self.build_url() + cluster_path
        payload = {
            "name": user_name,
        }
        resp = self.post_request(url, payload)
        return parse_one_user(resp["user"])

    def delete_database(self, cluster_id: str, database_name: str):
        cluster_path = f"/{cluster_id}/dbs/{database_name}"
        url = self.build_url() + cluster_path
        self.delete_request(url)
    
    def delete_database_user(self, cluster_id: str, username: str):
        cluster_path = f"/{cluster_id}/users/{username}"
        url = self.build_url() + cluster_path
        self.delete_request(url)

    def get_database_user(self, cluster_id: str, username: str) -> DatabaseUser:
        cluster_path = f"/{cluster_id}/users/{username}"
        url = self.build_url() + cluster_path
        resp = self.get_request(url)
        return parse_one_user(resp["user"])



import boto3
from botocore.exceptions import ClientError

from edos.exceptions import SpaceCreationError
from edos.settings import conf


class SpacesService:
    def __init__(self):
        session = boto3.session.Session()
        self.client = session.client(
            "s3",
            region_name=conf.AWS_REGION,
            endpoint_url=conf.AWS_ENDPOINT_URL,
            aws_access_key_id=conf.AWS_CONFIG.get("aws_access_key"),
            aws_secret_access_key=conf.AWS_CONFIG.get("aws_secret_key"),
        )

    def get_spaces(self) -> list[str]:
        response = self.client.list_buckets()
        return [space["Name"] for space in response["Buckets"]]

    def create_space(self, name: str):
        try:
            self.client.create_bucket(Bucket=name)
        except ClientError as e:
            raise SpaceCreationError(str(e))



import abc
import json

from httpx import request

from edos.api.constants import do_base_url
from edos.exceptions import UserReadableException
from edos.settings import conf


class DigitalOceanApi(abc.ABC):
    path = ""

    def __init__(self, token=None):
        try:
            self._token = token if token else conf.DIGITAL_OCEAN_CONFIG["token"]
        except (AttributeError, TypeError) as ex:
            raise UserReadableException("Could not find DO token in {}".format(conf.CONFIG_PATH)) from ex

    def create_request(self, method, url, headers, payload=None):
        try:
            response = request(method=method, url=url, headers=headers, data=payload)
            response.raise_for_status()
        except Exception as ex:
            if conf.INTERACTIVE:
                raise UserReadableException("DO API {} Request Failed: {})".format(method, url) + str(ex)) from ex
            raise ex
        if response.status_code == 204:
            return {}
        return response.json()

    def get_request(self, url, headers=None):
        if not headers:
            headers = {
                "Authorization": "Bearer " + self._token,
            }
        return self.create_request("GET", url, headers)

    def post_request(self, url, payload, headers=None):
        if not headers:
            headers = {
                "Accept": "application/json",
                "Content-Type": "application/json",
                "Authorization": "Bearer " + self._token,
            }
        return self.create_request("POST", url, headers, json.dumps(payload))

    def delete_request(self, url, headers=None):
        if not headers:
            headers = {
                "Authorization": "Bearer " + self._token,
            }
        return self.create_request("DELETE", url, headers)

    def build_url(self):
        return do_base_url + self.path



import logging

import click

from edos.services.spaces_service import SpacesService

LOG = logging.getLogger(__name__)


@click.group()
def spaces():
    """Digital Ocean Spaces (object storage) management"""


@spaces.command()
def ls():
    """List all spaces"""
    for space in SpacesService().get_spaces():
        click.echo(space)


@spaces.command()
@click.argument("name")
def create(name):
    """Create a space"""
    if not click.confirm(f'Creating space "{name}", proceed?'):
        click.echo("creation cancelled")
        return
    SpacesService().create_space(name)
    click.echo(click.style("Created space with name: ", fg="green") + name)



import logging
import shlex
import tempfile

import httpx
from edos.api.digitalocean.databases.api import DatabaseApi
from edos.api.digitalocean.databases.models.cluster import DatabaseCluster
from edos.api.digitalocean.databases.models.database import Database
from edos.api.digitalocean.databases.models.database_user import DatabaseUser
from edos.exceptions import DatabaseCreationError, DatabaseEngineNotSupported
from edos.settings import conf
import subprocess

logger = logging.getLogger(__name__)

class DatabaseService:
    def __init__(self):
        self.api = DatabaseApi()
        self.db_users_cache = {}

    def get_clusters(self) -> dict[str, str]:
        """
        :return: {cluster_id, cluster_name}
        """
        clusters = self.api.get_clusters()
        res = {}
        for cluster in clusters:
            res[cluster.id] = cluster.name
        return res

    def get_cluster(self, cluster_id: str) -> DatabaseCluster:
        if cluster_id == "psql":
            return self.api.get_one_cluster(conf.PSQL_CLUSTER_ID)
        if cluster_id == "psql-dev":
            return self.api.get_one_cluster(conf.PSQL_DEV_CLUSTER_ID)
        if cluster_id == "mysql":
            return self.api.get_one_cluster(conf.MYSQL_CLUSTER_ID)
        return self.api.get_one_cluster(cluster_id)

    def get_databases(self, cluster_id: str) -> list[str]:
        """
        :param cluster_id: cluster where is database stored
        :return: list of database names
        """
        databases = self.api.get_databases(cluster_id)
        return [db.name for db in databases]

    def get_database(self, cluster_id: str, database_name: str) -> Database:
        return self.api.get_database(cluster_id, database_name)
    
    def get_connection_string(self, cluster: DatabaseCluster, database: Database, user: DatabaseUser) -> str:
        if cluster.engine == "pg":
            prefix = "postgresql://"
        elif cluster.engine == "mysql":
            prefix = "mysql://"
        else:
            raise DatabaseEngineNotSupported("Bad cluster. Only PG and Mysql clusters are supported")
        
        private_connection = cluster.privateConnection
        return (
            f"{prefix}{user.name}:{user.password}@{private_connection.host}:"
            f"{private_connection.port}/{database.name}"
        )

    def create_database(self, cluster_id: str, project_name: str, from_dump_file: str | None = None, existing_user: str | None = None, upsert: bool = False) -> str:
        """
        this command creates a database and user

        so this command will be doing this:
            - creates a database
            - creates a database user
            - get cluster
            - create db_url from private connection from cluster
                and database user/password

        In case, that create user fail, you must delete database, for consistency

        :param cluster_id: cluster where database will be stored
        :param project_name: name of database and user
        :return: database url for secrets
        """
        is_database_fresh = True
        cluster = self.get_cluster(cluster_id)
        if from_dump_file and cluster.engine != "pg":
            raise DatabaseEngineNotSupported("Restoring from dump file is only supported for PG clusters")
        try:
            database = self.api.create_database(cluster.id, project_name)
        except httpx.HTTPStatusError as e:
            if upsert and e.response.status_code == 422:
                is_database_fresh = False
                database = self.api.get_database(cluster.id, project_name)
                existing_user = project_name  # assume user with the same name exists
            else:
                raise e
        if existing_user:
            user = self.get_database_user(cluster.id, existing_user)
            if not user:
                raise DatabaseCreationError(f"User {existing_user} does not exist in cluster {cluster.id}")
        else:
            try:
                user = self.api.create_database_user(cluster.id, project_name)
            except httpx.HTTPStatusError as e:
                logger.info(f"Database user creation failed: {e}, ignoring and proceeding")
        
        if cluster.engine == "pg":
            # Grant all privileges to the new user on the new database
            grant_db_cmd = f'GRANT ALL ON DATABASE "{database.name}" TO "{user.name}";'
            grant_schema_cmd = f'GRANT ALL ON SCHEMA public TO "{user.name}";'
            self.run_psql_shell(cluster, "defaultdb", "doadmin", input_string=grant_db_cmd)
            self.run_psql_shell(cluster, database.name, "doadmin", input_string=grant_schema_cmd)

            # import dump file if provided, but only if it's a newly created database
            if from_dump_file and is_database_fresh:
                self.run_psql_shell(cluster, database.name, user.name, input_filename=from_dump_file)

        return self.get_connection_string(cluster, database, user)

    def delete_database(self, cluster_id: str, db_name: str, db_user: str | None = None):
        """
        this command deletes a database and user

        :param cluster_id: cluster where is database stored
        :param db_name: name of database
        :param db_user: name of user, optional
        :return: None
        """
        cluster = self.get_cluster(cluster_id)
        try:
            self.api.delete_database(cluster.id, db_name)
            logger.info(f"Database {db_name} deleted successfully from cluster {cluster.id}")
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                logger.warning(f"Database {db_name} does not exist in cluster {cluster.id}, cannot delete")
            else:
                raise e
        if db_user:
            try:
                self.api.delete_database_user(cluster.id, db_user)
                logger.info(f"User {db_user} deleted successfully from cluster {cluster.id}")
            except httpx.HTTPStatusError as e:
                if e.response.status_code == 404:
                    logger.warning(f"User {db_user} does not exist in cluster {cluster.id}, cannot delete")
                else:
                    raise e

    def get_database_user(self, cluster_id, username) -> DatabaseUser:
        if (cluster_id, username) not in self.db_users_cache:
            self.db_users_cache[(cluster_id, username)] = self.api.get_database_user(cluster_id, username)
        return self.db_users_cache[(cluster_id, username)]

    def run_psql_shell(self, cluster: DatabaseCluster, database_name: str, username: str, input_filename: str | None = None, input_string: str | None = None):
        if input_filename and input_string:
            raise ValueError("You can use only one of input_filename or input_string")
        user = self.get_database_user(cluster.id, username)
        private_connection = cluster.privateConnection
        psql_cmd = (
            f'psql --host={private_connection.host} '
            f'--port={private_connection.port} --username={user.name} --dbname={database_name}'
        )
        ssh_cmd = [
            "ssh", conf.SSH_CONNECTION,
            f"PGPASSWORD='{user.password}' {psql_cmd}"
        ]

        logger.info(f"Running psql with flags: {psql_cmd}")

        if input_filename:
            # Use shell to handle input redirection
            ssh_cmd_shell = (
                f"ssh {conf.SSH_CONNECTION} PGPASSWORD='{user.password}' {psql_cmd} < \"{input_filename}\""
            )
            subprocess.run(ssh_cmd_shell, shell=True)
        elif input_string:
            # Pipe the input_string to psql via tempfile
            with tempfile.NamedTemporaryFile(delete=False) as temp_file:
                temp_file.write(input_string.encode())
                temp_file.close()
                ssh_cmd_shell = (
                    f"ssh {conf.SSH_CONNECTION} PGPASSWORD='{user.password}' {psql_cmd} < \"{temp_file.name}\""
                )
                subprocess.run(ssh_cmd_shell, shell=True)
        else:
            subprocess.run(ssh_cmd, shell=False)

    # def recreate_database(self, cluster_id: str, db_name: str):
    #     """
    #     this command just deletes database and create it again
    #     (when it needs to be deleted like uploading a dump)
    #     :param cluster_id: cluster where is database stored
    #     :param db_name: name of database
    #     :return: None - because db_url is the same
    #     """
    #     self.api.delete_database(cluster_id, db_name)
    #     self.api.create_database(cluster_id, db_name)
